#include <zephyr/kernel.h>
#include <stdio.h>
#include <string.h>
#include <drv/grspw.h>
#include <drv/gr716/grspw.h>
#include "grspw_pkt_lib.h"

#include <zephyr/shell/shell.h>

#define UNUSED(v) (void) sizeof (v)
#define DBG(fmt...) printk(fmt)

//#define PRINT_GRSPW_RESET_CFG

static int nospw = 0;

/* SpaceWire parameters */
#define SPW_PROT_ID 155

#define PKT_SIZE 32

struct grspw_device {
  struct grspw_dev dev;

  /* Test structures */
  struct grspw_config cfg;
  int run;

  /* RX and TX lists with packet buffers */
  struct grspw_list rx_list, tx_list, tx_buf_list;
  int rx_list_cnt, tx_list_cnt, tx_buf_list_cnt;
};

#define DEV(device) ((struct grspw_dev *)(device))

#define DEVS_MAX 2
static struct grspw_device devs[DEVS_MAX];

static struct grspw_config dev_def_cfg =
{
    .adrcfg =
    {
      .promiscuous = 1, /* Detect all packets */
      .def_addr = 0x40, /* updated in dev_init() */
      .def_mask = 0,
      .dma_nacfg =
      {
        /* Since only one DMA Channel is used, only
         * the default Address|Mask is used.
         */
        {
          .node_en = 0,
          .node_addr = 0,
          .node_mask = 0,
        },
        {
          .node_en = 0,
          .node_addr = 0,
          .node_mask = 0,
        },
        {
          .node_en = 0,
          .node_addr = 0,
          .node_mask = 0,
        },
        {
          .node_en = 0,
          .node_addr = 0,
          .node_mask = 0,
        },
      },
    },
    .rmap_cfg = 0,    /* Disable RMAP */
    .rmap_dstkey = 0, /* No RMAP DESTKEY needed when disabled */
    .enable_chan_mask = 1,  /* Enable only the first DMA Channel */
    .chan =
    {
      {
        .flags = DMAFLAG_NO_SPILL,
        .rxmaxlen = PKT_SIZE+4,
      },
      /* The other 3 DMA Channels are unused */
    },
};

/* SpaceWire Routing table entry */
struct route_entry {
  unsigned char dstadr[16]; /* 0 terminates array */
};

/* SpaceWire packet payload (data) content layout */
struct pkt_hdr {
  unsigned char addr;
  unsigned char protid;
  unsigned char port_src; /* port index of source */
  unsigned char resv2; /* Zero for now */
  unsigned int data[(PKT_SIZE-4)/4];
};

struct spwpkt {
  struct grspw_pkt p;
  unsigned long long data[PKT_SIZE/8+1]; /* 32 bytes of data - 4byte data-header (8 extra bytes to avoid truncated bad packets)*/
  unsigned long long hdr[2]; /* up to 16byte header (path address) */
};

/* All packet buffers used by application */
// #define PACKET_BUFFERS_PER_DEV (16+120)
#define PACKET_BUFFERS_PER_DEV (16+30)
static struct spwpkt pkts[DEVS_MAX][PACKET_BUFFERS_PER_DEV];

/* Initialize packet header, and Data payload header */
static void pkt_init_hdr(struct grspw_pkt *pkt, struct route_entry *route, int idx)
{
  int i;
  struct pkt_hdr *pkt_hdr = (struct pkt_hdr *)pkt->data;
  unsigned char *hdr = pkt->hdr;

  /* If path addressing we put non-first Destination Addresses in
   * header. route->dstadr[0] is always non-zero.
   */
  i = 0;
  while ( route->dstadr[i+1] != 0 ) {
    hdr[i] = route->dstadr[i];
    i++;
  }
  /* Put last address in pkthdr->addr */
  pkt->hlen = i;
  pkt_hdr->addr = route->dstadr[i];
  pkt_hdr->protid = SPW_PROT_ID;
  pkt_hdr->port_src = idx;
  pkt_hdr->resv2 = 0;
}

#ifndef NELEM
#define NELEM(a) ((int) (sizeof(a) / sizeof(a[0])))
#endif

static void init_pkts(void)
{
  struct spwpkt *pkt;
  int i, j;

  memset(&pkts[0][0], 0, sizeof(pkts));

  for (i = 0; i < DEVS_MAX; i++) {
    grspw_list_clr(&devs[i].rx_list);
    grspw_list_clr(&devs[i].tx_list);
    grspw_list_clr(&devs[i].tx_buf_list);
    devs[i].rx_list_cnt = 0;
    devs[i].tx_list_cnt = 0;
    devs[i].tx_buf_list_cnt = 0;
    for (j = 0, pkt = &pkts[i][0]; j < NELEM(pkts[0]); j++, pkt = &pkts[i][j]) {
      pkt->p.pkt_id = (i << 8)+ j; /* unused */
      pkt->p.data = &pkt->data[0];
      pkt->p.hdr = &pkt->hdr[0];
      if (j < (NELEM(pkts[0])-8)) {
        /* RX buffer */

        /* Add to device RX list */
        grspw_list_append(&devs[i].rx_list, &pkt->p);
        devs[i].rx_list_cnt++;
      } else {
        /* TX buffer */
        pkt->p.dlen = PKT_SIZE;
        uint8_t *d = pkt->p.data;
        memset(d+4, i, PKT_SIZE-4);

        /* Add to device TX list */
        grspw_list_append(&devs[i].tx_buf_list, &pkt->p);
        devs[i].tx_buf_list_cnt++;
      }
    }
  }
}

static int dev_init(int idx)
{
  struct grspw_device *devi = &devs[idx];
  struct grspw_dev    *dev = &devi ->dev;
  int i, ctrl, clkdiv;

  printf(" Initializing SpaceWire device %d\n", idx);

  memset(dev, 0, sizeof(*devi));

  dev->index = idx;
  dev->dh = grspw_open(idx);
  if (dev->dh == NULL) {
    printf("Failed to open GRSPW device %d\n", idx);
    return -1;
  }
  grspw_hw_support(dev->dh, &dev->hwsup);
#ifdef PRINT_GRSPW_RESET_CFG
  grspw_config_read(dev, &devi->cfg);
  printf("\n\n---- DEFAULT CONFIGURATION FROM DRIVER/HARDWARE ----\n");
  grspw_cfg_print(&dev->hwsup, &devi->cfg);
#endif
  devi->cfg = dev_def_cfg;
  devi->cfg.adrcfg.def_addr = 0x40 + idx;

  if (grspw_cfg_set(dev, &devi->cfg)) {
    grspw_close(dev->dh);
    return -1;
  }
#ifdef PRINT_GRSPW_RESET_CFG
  printf("\n\n---- APPLICATION CONFIGURATION ----\n");
  grspw_cfg_print(&dev->hwsup, &devi->cfg);
  printf("\n\n");
#endif

  /* This will result in an error if only one port available */
  if (dev->hwsup.nports < 2) {
    int port = 1;
    if ( grspw_port_ctrl(dev->dh, &port) == 0 ) {
      printf("Succeeded to select port1, however only one PORT on dev %d!\n", dev->index);
      return -1;
    }
  }

  /* Try to bring link up at clockdiv = 3 but do not touch
   * start-up clockdivisor */
  clkdiv = -1;
  grspw_link_ctrl(dev->dh, NULL, NULL, &clkdiv);
  ctrl = LINKOPTS_ENABLE | LINKOPTS_AUTOSTART | LINKOPTS_START;
  clkdiv &= 0xff00;
  clkdiv |= 0x0003;
  grspw_link_ctrl(dev->dh, &ctrl, NULL, &clkdiv);
  #define GAISLER_SPW2_DMA 0x08a

  if ( (dev->hwsup.hw_version >> 16) == GAISLER_SPW2_DMA ) {
    printf(" NOTE: running on SPW-ROUTER DMA SpaceWire link (no link-state available)\n");
  } else {
    printf(" After Link Start:");
    printf(" %s\n", grspw_linkstate_to_name(grspw_link_state(dev->dh)));
  }
  devi->run = 0;

  for (i=0; i<dev->hwsup.ndma_chans; i++) {
    if (dev->dma[i])
      grspw_dma_stats_clr(dev->dma[i]);
  }
  
  grspw_list_clr(&devi->rx_list);
  grspw_list_clr(&devi->tx_list);
  grspw_list_clr(&devi->tx_buf_list);
  devi->rx_list_cnt = devi->tx_list_cnt = devi->tx_buf_list_cnt = 0;

  return 0;
}

int main(void) {
  int i;

  if (1) {
    grspw_init(GR716_GRSPW_DRV_ALL);
  }

  nospw = grspw_dev_count();
  if (nospw < 1) {
    printf("Found no SpaceWire cores, aborting\n");
    exit(0);
  }
  printf("Found %d SpaceWire cores\n", nospw);
  if (nospw > DEVS_MAX) {
    printf("Limiting to %d SpaceWire devices\n", DEVS_MAX);
    nospw = DEVS_MAX;
  }

  memset(devs, 0, sizeof(devs));
  for (i=0; i<nospw; i++) {
    if (dev_init(i)) {
      printf("Failed to initialize GRSPW%d\n", i);
      exit(0);
    }
    puts("");
    fflush(NULL);
  }

  /* Initialize GRSPW */
  init_pkts();

  printf("Starting SpW DMA channels\n");
  for (i = 0; i < nospw; i++) {
    printf("Starting GRSPW%d: ", i);
    fflush(NULL);
    if (grspw_start(DEV(&devs[i]))) {
      printf("Failed to initialize GRSPW%d\n", i);
      exit(0);
    }
    printf("DMA Started Successfully\n");
  }
  return 0;
}

static int dma_process_tx(struct grspw_device *dev);
static int dma_process_rx(struct grspw_device *dev);

enum {
  DEVNO_NONE      = -1,
  DEVNO_BAD       = -2,
};

static int get_devno(size_t argc, char **argv) {
  int devno = DEVNO_NONE;
  if (1 < argc) {
    char *endptr;
    devno = strtoul(argv[1], &endptr, 0);
    if (endptr == argv[1]) {
      devno = DEVNO_BAD;
    } else if (devno < 0 || devno >= nospw) {
      devno = DEVNO_BAD;
    }
  }
  return devno;
}

static int cmd_speed(const struct shell *sh, size_t argc, char **argv) {
  int devno;
  int ret;

  if (argc < 2) {
    shell_error(sh, "%s: command need device number", argv[0]);
    return 0;
  }

  devno = get_devno(argc, argv);
  if (devno < 0) {
    shell_error(sh, " Invalid device number");
    return 0;
  }
  if (argc < 3) {
    shell_error(sh, "%s: command needs RUN parameter", argv[0]);
    return 0;
  }

  char *endptr;
  unsigned int rundiv = strtoul(argv[2], &endptr, 0);
  if (endptr == argv[2]) {
    shell_error(sh, "%s: error parsing RUN parameter %s", argv[0], argv[2]);
    return 0;
  }
  unsigned int startdiv = 0;

  if (3 < argc) {
    startdiv = strtoul(argv[3], &endptr, 0);
    if (endptr == argv[3]) {
      shell_error(sh, "%s: error parsing START parameter %s", argv[0], argv[3]);
      return 0;
    }
  }

  struct grspw_priv *dev = devs[devno].dev.dh;
  ret = grspw_set_clkdiv(dev, rundiv | startdiv << 8);
  if (ret) {
    shell_warn(sh, "ERROR: grspw_link_ctrl() says '%d'", ret);
  } else {
    shell_info(sh, "Successfully set RUN:%u, START:%u", rundiv, startdiv);
  }
  return 0;
}

static int cmd_link(const struct shell *sh, size_t argc, char **argv) {
  int ntotal = nospw;
  int devno;

  devno = get_devno(argc, argv);
  if (devno == DEVNO_BAD) {
    shell_error(sh, " Invalid device number\n");
    return 1;
  } else if (devno == DEVNO_NONE) {
    devno = 0;
  } else {
    ntotal = 1;
  }
  for (int i = 0; i < ntotal; i++) {
    struct grspw_link_state state;
    grspw_link_state_get(DEV(&devs[devno]), &state);
    shell_info(sh, " GRSPW%d link/port setup:", devno);
    grspw_linkstate_print(&state);
    shell_info(sh, "");
    devno++;
  }
  return 0;
}

static int cmd_r(const struct shell *sh, size_t argc, char **argv) {
  int ntotal = nospw;
  int devno;

  devno = get_devno(argc, argv);
  if (devno == DEVNO_BAD) {
    shell_error(sh, " Invalid device number\n");
    return 1;
  } else if (devno == DEVNO_NONE) {
    devno = 0;
  } else {
    ntotal = 1;
  }
  for (int i = 0; i < ntotal; i++) {
    shell_info(sh, " grspw%d: --- receive begin ---", devno);
    dma_process_rx(&devs[devno]);
    shell_info(sh, " grspw%d: --- receive end   ---", devno);
    devno++;
  }
  return 0;
}

static int cmd_x(const struct shell *sh, size_t argc, char **argv) {
  int devno;

  if (argc < 2) {
    shell_error(sh, "%s: command need device number", argv[0]);
    return 0;
  }

  devno = get_devno(argc, argv);
  if (devno < 0) {
    shell_error(sh, " Invalid device number");
    return 0;
  }

  if (argc > 17 || argc < 3) {
    shell_error(sh, " Invalid routing path");
    return 1;
  }

  struct grspw_pkt *pkt;
  struct route_entry route;
  memset(&route, 0, sizeof(route));
  for (size_t i = 2; i < argc; i++) {
    char *endptr;
    route.dstadr[i - 2] = strtoul(argv[i], &endptr, 0);
    if (endptr == argv[i]) {
      shell_error(sh, "%s: error parsing PATH or NODE (%s)", argv[0], argv[i]);
      return 0;
    }
  }

  pkt = grspw_list_remove_head(&devs[devno].tx_buf_list);
  if (pkt == NULL) {
    printf(" No free transmit buffers available\n");
    return 1;
  }
  devs[devno].tx_buf_list_cnt--;
  pkt_init_hdr(pkt, &route, devno);

  shell_info(sh, " X%d: scheduling packet on GRSPW%d", devno, devno);
  /* Send packet by adding it to the tx_list */
  grspw_list_append(&devs[devno].tx_list, pkt);
  devs[devno].tx_list_cnt++;
  dma_process_tx(&devs[devno]);
  return 0;
}

static int cmd_stats(const struct shell *sh, size_t argc, char **argv) {
  int ntotal = nospw;
  int devno;

  devno = get_devno(argc, argv);
  if (devno == DEVNO_BAD) {
    shell_error(sh, " Invalid device number\n");
    return 1;
  } else if (devno == DEVNO_NONE) {
    devno = 0;
  } else {
    ntotal = 1;
  }

  for (int i = 0; i < ntotal; i++) {
    struct grspw_stats stats;
    int tx_hwcnt, rx_hwcnt;

    shell_info(sh, " grspw%d: --- info begin ---", devno);
    /* Print statistics */
    shell_info(sh, "\n\n--- GRSPW%d Driver Stats ---", devno);
    /* NOTE: only on the first dma channel */
    rx_hwcnt = grspw_dma_rx_count(devs[devno].dev.dma[0]);
    tx_hwcnt = grspw_dma_tx_count(devs[devno].dev.dma[0]);
    shell_info(sh, " DRVQ RX_HWCNT: %d", rx_hwcnt);
    shell_info(sh, " DRVQ TX_HWCNT: %d", tx_hwcnt);

    grspw_stats_get(DEV(&devs[devno]), &stats);
    grspw_stats_print(DEV(&devs[devno]), &stats);
    shell_info(sh, " grspw%d: --- info end   ---", devno);
    devno++;
  }
  return 0;
}

static int cmd_info(const struct shell *sh, size_t argc, char **argv) {
  int ntotal = nospw;
  int devno;

  devno = get_devno(argc, argv);
  if (devno == DEVNO_BAD) {
    shell_error(sh, " Invalid device number\n");
    return 1;
  } else if (devno == DEVNO_NONE) {
    devno = 0;
  } else {
    ntotal = 1;
  }
  for (int i = 0; i < ntotal; i++) {
    shell_info(sh, " grspw%d: --- info begin ---", devno);
    grspw_cfg_print(&devs[devno].dev.hwsup, NULL);
    shell_info(sh, " grspw%d: --- info end   ---\n", devno);
    devno++;
  }
  return 0;
}

static int dma_process_rx(struct grspw_device *dev)
{
  int cnt;
  struct grspw_list lst;
  struct grspw_pkt *pkt;
  unsigned char *c;

  /* Prepare receiver with packet buffers */
  if (dev->rx_list_cnt > 0) {
    cnt = grspw_dma_rx_prepare(dev->dev.dma[0], &dev->rx_list);
    if (cnt < 0) {
      printf("rx_prep failed %d\n", cnt);
      return -1;
    }
    if (1) {
      printf("GRSPW%d: Prepared %d RX packet buffers for future "
       "reception\n", dev->dev.index, dev->rx_list_cnt);
    }
    dev->rx_list_cnt -= cnt;
  }

  /* Try to receive packets on receiver interface */
  grspw_list_clr(&lst);
  cnt = grspw_dma_rx_recv(dev->dev.dma[0], &lst);
  if (cnt < 0) {
    printf("rx_recv failed %d\n", cnt);
    return -1;
  }
  if (cnt > 0) {
    printf("GRSPW%d: Recevied %d packets\n", dev->dev.index, cnt);
    for (pkt = lst.head; pkt; pkt = pkt->next) {
      if ((pkt->flags & RXPKT_FLAG_RX) == 0) {
        printf(" PKT not received.. buf ret\n");
        continue;
      } else if (pkt->flags & (RXPKT_FLAG_EEOP | RXPKT_FLAG_TRUNK)) {
        printf(" PKT RX errors:");
        if (pkt->flags & RXPKT_FLAG_TRUNK) {
          printf(" truncated");
        }
        if (pkt->flags & RXPKT_FLAG_EEOP) {
          printf(" EEP");
        }
        printf(" (0x%x)", pkt->flags);
      } else {
        printf(" PKT");
      }
      c = (unsigned char *)pkt->data;
      printf(" of length %d bytes, 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x...\n",
       (int) pkt->dlen, c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]);
    }

    /* Reuse packet buffers by moving packets to rx_list */
    grspw_list_append_list(&dev->rx_list, &lst);
    dev->rx_list_cnt += cnt;
  }
  return 0;
}

static int dma_process_tx(struct grspw_device *dev)
{
  int cnt;
  struct grspw_list lst;
  struct grspw_pkt *pkt;
  unsigned char *c;

  /*** TX PART ***/

  /* Reclaim already transmitted buffers */
  grspw_list_clr(&lst);
  cnt = grspw_dma_tx_reclaim(dev->dev.dma[0], &lst);
  if (cnt < 0 ) {
    printf("tx_reclaim failed %d\n", cnt);
    exit(0);
  }
  /* put sent packets in end of send queue for retransmission */
  if (cnt > 0) {
    /*printf("GRSPW%d: Reclaimed %d TX packet buffers\n",
      dev->index, cnt);*/
    /* Clear transmission flags */
    pkt = lst.head;
    while (pkt) {
      if ((pkt->flags & TXPKT_FLAG_TX) == 0) {
        printf(" One Packet TX failed\n");
      } else if (pkt->flags & TXPKT_FLAG_LINKERR) {
        printf(" One Packet with TX errors\n");
      }
      pkt = pkt->next;
    }

    grspw_list_append_list(&dev->tx_buf_list, &lst);
    dev->tx_buf_list_cnt += cnt;
  }

  /* Send packets in the tx_list queue */
  if (dev->tx_list_cnt > 0) {
    printf("GRSPW%d: Sending %d packets\n", dev->dev.index, dev->tx_list_cnt);
    for (pkt = dev->tx_list.head; pkt; pkt = pkt->next) {
      printf(" PKT of length %u bytes,", (unsigned) pkt->hlen+pkt->dlen);
      for (size_t i = 0; i < pkt->hlen+pkt->dlen && i < 8; i++) {
        if (i < pkt->hlen) {
          c = i + (unsigned char *)pkt->hdr;
        } else {
          c = i - pkt->hlen + (unsigned char *)pkt->data;
        }
        printf(" 0x%02x", *c);
      }
      printf("\n");
    }
    cnt = grspw_dma_tx_send(dev->dev.dma[0], &dev->tx_list);
    if (cnt < 0) {
      printf("tx_send failed %d\n", cnt);
      exit(0);
    }
    dev->tx_list_cnt -= cnt;
  }
  return 0;
}

#if 0
/* print examples */
shell_fprintf(sh, SHELL_INFO, "Print info message\n");
shell_print(sh, "Print simple text.");
shell_warn(sh, "Print warning text.");
shell_error(sh, "Print error text.");

static int cmd_demo_params(const struct shell *sh, size_t argc, char **argv) {
  int cnt;

  shell_print(sh, "argc = %d", argc);
  for (cnt = 0; cnt < argc; cnt++) {
    shell_print(sh, "  argv[%d] = %s", cnt, argv[cnt]);
  }
  return 0;
}
#endif

/* Create subcommands (level 1 command) for command "spw" */
SHELL_STATIC_SUBCMD_SET_CREATE(
  sub_spw,
  SHELL_CMD(link,     NULL, "Link and port set-up and state\n  link [DEV]",     cmd_link),
  SHELL_CMD(stats,    NULL, "Statistics\n  spw stats [DEV]",                    cmd_stats),
  SHELL_CMD(info,     NULL, "Get device information\n  spw info [DEV]",         cmd_info),
  SHELL_CMD(clkdiv,   NULL, "Set clock divisor\n  spw clkdiv DEV RUN [START]",  cmd_speed),
  SHELL_CMD(x,        NULL, "Transmit one packet\n  spw x DEV [PATH_0 .. PATH_N] NODE_ID", cmd_x),
  SHELL_CMD(r,        NULL, "Receive messages\n  spw r [DEV]",                  cmd_r),
  SHELL_SUBCMD_SET_END
);

/* create root (level 0) command "spw" without a handler */
SHELL_CMD_REGISTER(spw, &sub_spw, "GRSPW commands", NULL);

