#include <stdint.h>
#include <stdbool.h>
#include <drv/auto.h>
#include <drv/osal.h>
#include <drv/drvret.h>
#include <drv/regs/grspwrouter-regs.h>

/* Defines used to parse the status value from grspwrouter_port_status() */
#define GRSPWROUTER_STATUS_ERR_TRUNC      (1 << 29)
#define GRSPWROUTER_STATUS_ERR_INTTRUNC   (1 << 28)
#define GRSPWROUTER_STATUS_ERR_RMAP       (1 << 27)
#define GRSPWROUTER_STATUS_ERR_NOTRDY     (1 << 26)
#define GRSPWROUTER_STATUS_START_REQUEST  (1 << 22)
#define GRSPWROUTER_STATUS_SPILL          (1 << 21)
#define GRSPWROUTER_STATUS_ACTIVE_STATUS  (1 << 20)
#define GRSPWROUTER_STATUS_ACTIVE_PORT    (1 << 19)
#define GRSPWROUTER_STATUS_ERR_TIMEOUT    (1 << 18)
#define GRSPWROUTER_STATUS_ERR_MEM        (1 << 17)
#define GRSPWROUTER_STATUS_TX_FIFO_FULL   (1 << 16)
#define GRSPWROUTER_STATUS_RX_FIFO_EMPTY  (1 << 15)
#define GRSPWROUTER_STATUS_RX_BUSY        (1 <<  6)
#define GRSPWROUTER_STATUS_TX_BUSY        (1 <<  5)
#define GRSPWROUTER_STATUS_ERR_ADRS       (1 <<  4)
#define GRSPWROUTER_STATUS_ERR_CREDIT     (1 <<  3)
#define GRSPWROUTER_STATUS_ERR_ESCAPE     (1 <<  2)
#define GRSPWROUTER_STATUS_ERR_DISCON     (1 <<  1)
#define GRSPWROUTER_STATUS_ERR_PARITY     (1 <<  0)
#define GRSPWROUTER_STATUS_PORT_TYPE(status)  (status >> 30)
#define GRSPWROUTER_STATUS_LINK_STATE(status) ((status >> 12) &  0x7)
#define GRSPWROUTER_STATUS_INPUT_PORT(status) ((status >> 7 ) & 0x1f)

/* Defines used with grspwrouter_isr_register() to enable interrupt sources */
#define GRSPWROUTER_INTERRUPT_PNP        (1 << 10)
#define GRSPWROUTER_INTERRUPT_NOTRDY     (1 <<  9)
#define GRSPWROUTER_INTERRUPT_RUN_STATE  (1 <<  8)
#define GRSPWROUTER_INTERRUPT_INTTRUNC   (1 <<  7)
#define GRSPWROUTER_INTERRUPT_TRUNC      (1 <<  6)
#define GRSPWROUTER_INTERRUPT_TIMEOUT    (1 <<  5)
#define GRSPWROUTER_INTERRUPT_CONF_PORT  (1 <<  4)
#define GRSPWROUTER_INTERRUPT_RMAP       (1 <<  3)
#define GRSPWROUTER_INTERRUPT_ADRS       (1 <<  2)
#define GRSPWROUTER_INTERRUPT_LINK       (1 <<  1)
#define GRSPWROUTER_INTERRUPT_MEM        (1 <<  0)
#define GRSPWROUTER_INTERRUPT_ALL          (0x7ff)
#define GRSPWROUTER_INTERRUPT_NONE             (0)

#define GRSPWROUTER_IC_DIS_RX_INT               1
#define GRSPWROUTER_IC_DIS_TX_INT               2
#define GRSPWROUTER_IC_DIS_RX_ACK               4
#define GRSPWROUTER_IC_DIS_TX_ACK               8

#define GRSPWROUTER_PORT_SPW                    0
#define GRSPWROUTER_PORT_AMBA                   1
#define GRSPWROUTER_PORT_FIFO                   2
#define GRSPWROUTER_PORT_CUSTOM                 3

#define GRSPWROUTER_PORT(p)              (1 << p)

#define GRSPWROUTER_GR740_SPW_1         (1 <<  1)
#define GRSPWROUTER_GR740_SPW_2         (1 <<  2)
#define GRSPWROUTER_GR740_SPW_3         (1 <<  3)
#define GRSPWROUTER_GR740_SPW_4         (1 <<  4)
#define GRSPWROUTER_GR740_SPW_5         (1 <<  5)
#define GRSPWROUTER_GR740_SPW_6         (1 <<  6)
#define GRSPWROUTER_GR740_SPW_7         (1 <<  7)
#define GRSPWROUTER_GR740_SPW_8         (1 <<  8)
#define GRSPWROUTER_GR740_AMBA_0        (1 <<  9)
#define GRSPWROUTER_GR740_AMBA_1        (1 << 10)
#define GRSPWROUTER_GR740_AMBA_2        (1 << 11)
#define GRSPWROUTER_GR740_AMBA_3        (1 << 12)

#define GRSPWROUTER_ROUTE_PRIORITY              1
#define GRSPWROUTER_PACKET_DIST                 2

typedef enum link_state {
    GRSPWROUTER_LINK_ERROR_RESET = 0,
    GRSPWROUTER_LINK_ERROR_WAIT  = 1,
    GRSPWROUTER_LINK_READY       = 2,
    GRSPWROUTER_LINK_STARTED     = 3,
    GRSPWROUTER_LINK_CONNECTING  = 4,
    GRSPWROUTER_LINK_RUN_STATE   = 5
} link_state;

/**
 * @brief SpaceWire Router device callback function type
 *
 * A function of this type can be registered by ``grspwrouter_isr_register``
 * to be called by the ``grspwrouter_isr`` ISR when:
 *
 * * a packet for which this port was the input port has been spilled due to the packet length truncation feature (GRSPWROUTER_STATUS_ERR_TRUNC)
 * * a packet for which this port was the input port has been spilled due to the time-code / distributed interrupt code truncation feature (GRSPWROUTER_STATUS_ERR_INTTRUNC)
 * * an RMAP / SpaceWire Plug-and-Play command received on this port was spilled by the configuration port (GRSPWROUTER_STATUS_ERR_RMAP)
 * * a packet received on this port was spilled due to the spill-if-not-ready feature (GRSPWROUTER_STATUS_ERR_NOTRDY)
 * * a packet for which this port was the input port was spilled due to a packet timeout (GRSPWROUTER_STATUS_ERR_TIMEOUT)
 * * a memory error occur while accessing the on-chip memory in the ports (GRSPWROUTER_STATUS_ERR_MEM)
 * * an invalid address error occurred on this port (GRSPWROUTER_STATUS_ERR_ADRS)
 * * a credit error has occurred (GRSPWROUTER_STATUS_ERR_CREDIT)
 * * an escape error has occurred (GRSPWROUTER_STATUS_ERR_ESCAPE)
 * * a disconnect error has occurred (GRSPWROUTER_STATUS_ERR_DISCON)
 * * a parity error has occurred (GRSPWROUTER_STATUS_ERR_PARITY)
 * * a link enters runstate
 *
 * It will be called in the interrupt context.
 *
 * @param port The index of the port that triggered the interrupt
 * @param link_status The link status of the port that triggered the interrupt (Only for SpaceWire ports)
 * @param events Mask indicating the events that caused the interrupt to trigger
 * @param[in] isr_arg Custom pointer provided when the function was registered
 */

typedef void (*grspwrouter_isr_func)(uint8_t port, uint32_t link_status, uint32_t events, void *isr_arg);

typedef struct grspwrouter_dev {
    struct grspwrouter_regs* regs;
    uint32_t max_ports;
    uint32_t max_spw_ports;
    grspwrouter_isr_func isr;
    void * isr_arg;
    uint32_t interrupt;
    uint8_t open;
    struct osal_isr_ctx isr_ctx;
    uint32_t spw_freq;
} grspwrouter_dev;

typedef struct grspwrouter_devcfg {
    struct drv_devreg regs;
    struct grspwrouter_dev dev;
} grspwrouter_devcfg;

/**
 * @brief Register SpaceWire router devices using Plug-n-Play
 *
 * Registers any available SpaceWire router devices and returns
 * the number of devices found.
 *
 * @return The number of devices found and registered
 */
int grspwrouter_autoinit(void);

/**
 * @brief Manually register a single SpaceWire router device
 *
 * The configuration must include the location of the register
 * area and the interrupt number in ``devcfg->regs``. The
 * ``devcfg->dev`` member is used be the driver to store information.
 * The memory used by the ``devcfg`` argument must never be freed.
 *
 * @param[in] devcfg Settings defining the router device
 * @return DRV_OK on success
 */
drvret grspwrouter_register(struct grspwrouter_devcfg *devcfg);

/**
 * @brief Returns the number of SpaceWire routers registered to the driver
 *
 * @return The number of SpaceWire routers
 */
uint32_t grspwrouter_dev_count(void);

/**
 * @brief Closes a previously opened device
 *
 * The provided handle must have been previously opened by
 * ``grspwrouter_open()``.
 *
 * @param[in] dev A valid device handle
 * @retval DRV_OK on success
 * @retval DRV_INVAL if not previously opened by ``grspwrouter_open``
 */
drvret grspwrouter_close(grspwrouter_dev *dev);

/**
 * @brief Initialize handle to SpaceWire router driver
 *
 * This function returns a handle to SpaceWire router driver
 * for the device specified by ``index``.
 * 
 * The ``spw_freq`` argument shall specify the SpaceWire clock
 * frequency (in MHz) provided to the router. This value is used to
 * configure the initialization bit rate for the all the SpaceWire
 * links. It is also used by ``grspwrouter_port_link_start`` to
 * set the run state speed of individual links. Use the value 0
 * to keep the existing value.
 * 
 * The ``sys_freq`` arguments shall specify the system clock frequency
 * (in MHz). This value is used to configure the various timeout functionality
 * provided by the router. This function will set the timer scaler so
 * that all timers run at 10KHz. Use the value 0 to keep the existing value.
 * 
 * For the GR740 the default internal SpaceWire clock frequency is 400MHz. This
 * corresponds to an external clock frequency for a SPW_CLK of 50 MHz if the
 * default PLL configuration of 8x is used.
 *
 * @param index Index of the SpaceWire router device
 * @param spw_freq SpaceWire clock frequency
 * @param sys_freq System clock frequency
 *
 * @retval grspwrouter_dev on success
 * @retval NULL if no device with the provided index, or if already opened
 */
grspwrouter_dev * grspwrouter_open(uint32_t index, uint32_t spw_freq, int32_t sys_freq);

/**
 * @brief Start the SpaceWire link
 *
 * Configure the link rate to use and enable the link. The link rate shall
 * be specified in MBits/s.
 * 
 * This function can only be called on a SpaceWire port, not an AMBA port.
 *
 * @param[in] dev Valid router device handle
 * @param port Index of a valid SpaceWire port
 * @param link_rate The requested run-state link rate
 *
 * @retval DRV_OK on success
 * @retval DRV_INVAL port is not a SpaceWire port or invalid link rate
 */
drvret grspwrouter_port_link_start(grspwrouter_dev * dev, uint8_t port, uint32_t link_rate);

/**
 * @brief Stops the SpaceWire port link
 *
 * This function can only be called on a SpaceWire port, not an AMBA port.
 *
 * @param[in] dev Valid router device handle
 * @param port Index of a valid SpaceWire port
 *
 * @retval DRV_OK on success
 * @retval DRV_INVAL port is not a SpaceWire port
 */
drvret grspwrouter_port_link_stop(grspwrouter_dev * dev, uint8_t port);

/**
 * @brief Returns the link state of the SpaceWire port
 *
 * This function can only be called on a SpaceWire port, not an AMBA port.
 *
 * @param[in] dev Valid router device handle
 * @param port Index of a valid SpaceWire port
 * @param[out] status The current link state
 *
 * @retval DRV_OK on success
 * @retval DRV_INVAL port is not a SpaceWire port
 *
 */
drvret grspwrouter_port_link_status(grspwrouter_dev * dev, uint8_t port, link_state * status);

/**
 * @brief Return list of SpaceWire ports with links in runstate
 *
 * The mask returned by the function indicates which SpaceWire port links are in
 * runstate. Bit 1 is SpaceWire port 1, bit 2 is SpaceWire port 2, and so on.
 *
 * @param[in] dev Valid router device handle
 * @param[out] run_state Mask indicating runstate of each link
 *
 */
void grspwrouter_link_status(grspwrouter_dev *dev, uint32_t * run_state);

/**
 * @brief Return the status of the port
 *
 * This function returns the value of the status register for the port.
 *
 * The status value can be parsed using the following defines:
 *
 * * port type (SpaceWire/AMBA/FIFO/Custom) (GRSPWROUTER_STATUS_PORT_TYPE(status))
 * * a packet for which this port was the input port has been spilled due to the packet length truncation feature (GRSPWROUTER_STATUS_ERR_TRUNC)
 * * a packet for which this port was the input port has been spilled due to the time-code / distributed interrupt code truncation feature (GRSPWROUTER_STATUS_ERR_INTTRUNC)
 * * an RMAP / SpaceWire Plug-and-Play command received on this port was spilled by the configuration port (GRSPWROUTER_STATUS_ERR_RMAP)
 * * a packet received on this port was spilled due to the spill-if-not-ready feature (GRSPWROUTER_STATUS_ERR_NOTRDY)
 * * this port either was started, or currently is trying to start, due to the link-start-on-request feature (GRSPWROUTER_STATUS_START_REQUEST)
 * * a packet that is incoming on this port currently is being spilled (GRSPWROUTER_STATUS_SPILL)
 * * a packet arrives at this port and the port has been given access to the routing table (GRSPWROUTER_STATUS_ACTIVE_STATUS)
 * * the active SpaceWire ports if dual ports is implemented (GRSPWROUTER_STATUS_ACTIVE_PORT)
 * * a packet for which this port was the input port was spilled due to a packet timeout (GRSPWROUTER_STATUS_ERR_TIMEOUT)
 * * a memory error occur while accessing the on-chip memory in the ports (GRSPWROUTER_STATUS_ERR_MEM)
 * * transmit FIFO on this port is full (GRSPWROUTER_STATUS_TX_FIFO_FULL)
 * * receive FIFO on this port is empty (GRSPWROUTER_STATUS_RX_FIFO_EMPTY)
 * * current link state (GRSPWROUTER_STATUS_LINK_STATE(status))
 * * the number of the input port for the current or last packet transfer on this port (GRSPWROUTER_STATUS_INPUT_PORT(status))
 * * port is the input port of an ongoing packet transfer (GRSPWROUTER_STATUS_RX_BUSY)
 * * port is the output port of an ongoing packet transfer (GRSPWROUTER_STATUS_TX_BUSY)
 * * an invalid address error occurred on this port (GRSPWROUTER_STATUS_ERR_ADRS)
 * * a credit error has occurred (GRSPWROUTER_STATUS_ERR_CREDIT)
 * * an escape error has occurred (GRSPWROUTER_STATUS_ERR_ESCAPE)
 * * a disconnect error has occurred (GRSPWROUTER_STATUS_ERR_DISCON)
 * * a parity error has occurred (GRSPWROUTER_STATUS_ERR_PARITY)
 *
 * @param[in] dev Valid router device handle
 * @param port Index of a valid port
 * @param[out] status The port status register
 *
 * @retval DRV_OK on success
 * @retval DRV_INVAL port is not a valid port
  */
drvret grspwrouter_port_status(grspwrouter_dev * dev, uint8_t port, uint32_t *status);

/**
 * @brief Set up a route for incoming packets based on destination address
 *
 * Incoming packets with the destination address ``address`` will be routed
 * to the first available output port of the ones specified in the ``to_mask``.
 * If packet distribution has been enabled the same packet will be sent on all
 * specified output ports.
 *
 * The ``to_mask`` argument can be built using a mask where each bit index
 * corresponds to the SpaceWire port with the same index. The ``GRSPWROUTER_PORT()``
 * define can be used for this:
 *
 * ``to_mask = GRSPWROUTER_PORT(3) | GRSPWROUTER_PORT(4)``
 *
 * On the GR740 the following defines can be used:
 *
 * * AMBA port 0 (GRSPWROUTER_GR740_AMBA_0)
 * * AMBA port 1 (GRSPWROUTER_GR740_AMBA_1)
 * * AMBA port 2 (GRSPWROUTER_GR740_AMBA_2)
 * * AMBA port 3 (GRSPWROUTER_GR740_AMBA_3)
 * * SpaceWire port 1 (GRSPWROUTER_GR740_SPW_1)
 * * SpaceWire port 2 (GRSPWROUTER_GR740_SPW_2)
 * * SpaceWire port 3 (GRSPWROUTER_GR740_SPW_3)
 * * SpaceWire port 4 (GRSPWROUTER_GR740_SPW_4)
 * * SpaceWire port 5 (GRSPWROUTER_GR740_SPW_5)
 * * SpaceWire port 6 (GRSPWROUTER_GR740_SPW_6)
 * * SpaceWire port 7 (GRSPWROUTER_GR740_SPW_7)
 * * SpaceWire port 8 (GRSPWROUTER_GR740_SPW_8)
 *
 * Packets sent to the AMBA ports are handled by the SpaceWire driver.
 *
 * The router can be configured to automatically remove the first byte
 * of the packet, the byte that contains the destination address. This
 * is called header deletion.
 *
 * If the output port's link interface is not in run-state the router
 * can be ordered to wait until the link is up or to spill the packet.
 *
 * The ``options`` argument can be built by or:ing the following defines:
 *
 * * Set high priority when more than one packet is competing for the same output port (``GRSPWROUTER_ROUTE_PRIORITY``)
 * * Enable packet distribution (default group adaptive) (``GRSPWROUTER_PACKET_DIST``)
 *
 * @param[in] dev Valid router device handle
 * @param address Route incoming packets with this destination address
 * @param to_mask Route packets to these output ports
 * @param header_deletion Remove the first byte of the packet when routing it
 * @param spill_packet Spill the packet if the output port's link interface is not in run-state
 * @param options Enable high priority (GRSPWROUTER_ROUTE_PRIORITY) and/or packet distribution (GRSPWROUTER_PACKET_DIST)
 *
 * @retval DRV_OK on success
 * @retval DRV_INVAL if address is 0
 */
drvret grspwrouter_route_set(grspwrouter_dev *dev, uint8_t address, uint32_t to_mask, bool header_deletion, bool spill_packet, uint32_t options);

/**
 * @brief Disable a route for incoming packets based on destination address
 *
 * PrSTATUS the router from routing packets with a specific destination address.
 * Only logical addresses can be blocked. Packets with a physical destination address
 * will still be routed.
 *
 * @param[in] dev Valid router device handle
 * @param address Packets with this logical destination address will not be routed (32 - 255)
 *
 * @retval DRV_OK on success
 * @retval DRV_INVAL on non-logical address
 *
 */
drvret grspwrouter_route_disable(grspwrouter_dev *dev, uint8_t address);

/**
 * @brief Set up a static route for incoming packets on a specific port
 *
 * This function enables static routing for a port where incoming packets are
 * always routed unmodified to a specific output port regardless of the address
 * in the packet. By setting ``use_route_info`` to ``true`` it is
 * possible to use the normal route information to route the packet to multiple ports.
 *
 * @param[in] dev Valid router device handle
 * @param port Index of a valid port
 * @param destination Target port
 * @param use_route_info Use the target addresses configured by ``grspwrouter_route_set`` for the target
 *
 * @retval DRV_OK on success
 * @retval DRV_INVAL if static routing not supported, or invalid port or destination
 */
drvret grspwrouter_port_static_route_set(grspwrouter_dev * dev, uint8_t port, uint32_t destination, bool use_route_info);

/**
 * @brief Disable static routing for the port
 *
 * Disable static routing for the port.
 *
 * @param[in] dev Valid router device handle
 * @param port Index of a valid port
 *
 * @retval DRV_OK on success
 * @retval DRV_INVAL if invalid port
 */
drvret grspwrouter_port_static_route_disable(grspwrouter_dev * dev, uint8_t port);

/**
 * @brief Enable the handling of time codes
 *
 * Enable the router time code support. Also needs to be enabled for each port
 * that intend to use time codes using ``grspwrouter_port_tc_enable``.
 *
 * @param[in] dev Valid router device handle
 */
void grspwrouter_tc_enable(grspwrouter_dev *dev);

/**
 * @brief Disable time code support
 *
 * Disable the router time code support.
 *
 * @param[in] dev Valid router device handle
 */
void grspwrouter_tc_disable(grspwrouter_dev *dev);

/**
 * @brief Set the current time code to 0
 *
 * Sets the routers internal time representation to 0.
 *
 * @param[in] dev Valid router device handle
 */
void grspwrouter_tc_reset(grspwrouter_dev *dev);

/**
 * @brief Return the current time code
 *
 * Returns routers internal time representation.
 *
 * @param[in] dev Valid router device handle
 * @return Current time code
 */
uint8_t grspwrouter_tc_get(grspwrouter_dev *dev);

/**
 * @brief Enable time code support
 *
 * This function enables time codes to be sent and received via the port.
 * If ``router_time`` is ``true`` the router will not look at the timer value
 * and instead use its internal time representation.
 *
 * Time code support also needs to be enabled globally using ``grspwrouter_tc_enable``.
 *
 * @param[in] dev Valid router device handle
 * @param port Index of a valid port
 * @param router_time If true, always use the routers time, never the incoming time
 *
 * @retval DRV_OK on success
 * @retval DRV_INVAL if invalid port
 */
drvret grspwrouter_port_tc_enable(grspwrouter_dev * dev, uint8_t port, bool router_time);

/**
 * @brief Disable time code support
 *
 * Disables support for time codes for the port. Any time codes received will be dropped.
 *
 * @param[in] dev Valid router device handle
 * @param port Index of a valid SpaceWire port
 *
 * @retval DRV_OK on success
 * @retval DRV_INVAL if invalid port
 */
drvret grspwrouter_port_tc_disable(grspwrouter_dev * dev, uint8_t port);

/**
 * @brief Enable interrupt code support for port
 *
 * By default forwarding of both interrupt codes and interrupt acknowledgement codes in
 * both direction are enabled. This can be changed by or:ing the defines below together
 * to form an ``options`` argument:
 *
 * * Disable the transmission of interrupt codes (GRSPWROUTER_IC_DIS_TX_INT)
 * * Disable the reception of interrupt codes (GRSPWROUTER_IC_DIS_RX_INT)
 * * Disable the transmission of interrupt acknowledgement codes (GRSPWROUTER_IC_DIS_TX_ACK)
 * * Disable the reception of interrupt acknowledgement codes (GRSPWROUTER_IC_DIS_RX_ACK)
 *
 * Interrupt code support also needs to be enabled globally using ``grspwrouter_ic_enable``.
 *
 * @param[in] dev Valid router device handle
 * @param port Index of a valid SpaceWire port
 * @param options Options mask
 *
 * @retval DRV_OK on success
 * @retval DRV_INVAL if invalid port
 */
drvret grspwrouter_port_ic_enable(grspwrouter_dev * dev, uint8_t port, uint32_t options);

/**
 * @brief Disable interrupt code support for port
 *
 * Disables support for interrupt codes for the port. Any interrupt codes received will be dropped.
 *
 * @param[in] dev Valid router device handle
 * @param port Index of a valid SpaceWire port
 *
 * @retval DRV_OK on success
 * @retval DRV_INVAL if invalid port
 */
drvret grspwrouter_port_ic_disable(grspwrouter_dev * dev, uint8_t port);

/**
 * @brief Enable interrupt code support
 *
 * Enable the router interrupt code support. Also needs to be enabled for each port
 * that intend to send or receive interrupt codes using ``grspwrouter_port_ic_enable``.
 *
 * A timer can be configured that will trigger an interrupt when an
 * acknowledge reply is not received within the specified time period (100μs - 6.5s).
 *
 * A cooldown period can be configured that prevents new interrupts from
 * being submitted until the specified time period has passed (100μs - 3.1ms).
 *
 * Set the timeout to zero to disable.
 *
 * @param[in] dev Valid router device handle
 * @param timeout Timeout in microseconds (or 0 to disable) (100 - 6553500 in steps of 100)
 * @param cooldown Cooldown period in microseconds (or 0 to disable) (100 - 3100 in steps of 100)
 *
 * @retval DRV_OK on success
 * @retval DRV_INVAL if the timeout or cooldown value is too big
 */
drvret grspwrouter_ic_enable(grspwrouter_dev *dev, uint32_t timeout, uint32_t cooldown);

/**
 * @brief Disable interrupt code support
 *
 * Disable interrupt code support for all ports in router
 *
 * @param[in] dev Valid router device handle
 */
void grspwrouter_ic_disable(grspwrouter_dev *dev);

/**
 * @brief Register handler for port events
 *
 * Register a handler for the selected interrupt types. The defines below can be or:ed together
 * to form the ``mask`` argument:
 * 
 * * Generate an interrupt when a SpaceWire Plug and Play error has been detected in the configuration port (GRSPWROUTER_INTERRUPT_CONF_PNP)
 * * Generate an interrupt when a packet has been spilled because of the spill-if-not-ready feature (GRSPWROUTER_INTERRUPT_NOTRDY)
 * * Generate an interrupt when a SpaceWire link enters run-state (GRSPWROUTER_INTERRUPT_RUN_STATE)
 * * Generate an interrupt when a packet has been spilled because of the time code / distributed interrupt code truncation feature (GRSPWROUTER_INTERRUPT_INTTRUNC)
 * * Generate an interrupt when a packet has been spilled due to the packet length truncation feature (GRSPWROUTER_INTERRUPT_TRUNC)
 * * Generate an interrupt when a packet has been spilled due to the timeout mechanism (GRSPWROUTER_INTERRUPT_TIMEOUT)
 * * Generate an interrupt when either a header CRC error, protocol ID error, packet type error, early EOP, or early EEP has been detected in the configuration port (GRSPWROUTER_INTERRUPT_CONF_PORT)
 * * Generate an interrupt when an error has been detected in the configuration port for an RMAP command such that the PSTS.EC field is set to a non-zero value (GRSPWROUTER_INTERRUPT_CONF_RMAP)
 * * Generate an interrupt when an invalid address error has occurred on a port (GRSPWROUTER_INTERRUPT_ADRS)
 * * Generate an interrupt when a link error (parity, escape, credit, disconnect) has been detected on a SpaceWire port (GRSPWROUTER_INTERRUPT_LINK)
 * * Generate an interrupt when a memory error occur in any of the router's on-chip memories (GRSPWROUTER_INTERRUPT_MEM)
 *
 * The define `GRSPWROUTER_INTERRUPT_ALL` can be used to enable all interrupt types and `GRSPWROUTER_INTERRUPT_NONE` to disable all interrupt types.
 * 
 * @param[in] dev Valid router device handle
 * @param err_mask Interrupts that the handler should trigger on
 * @param port_mask Ports that the interrupts can be generated for
 * @param[in] isr Interrupt handler function pointer
 * @param[in] arg Custom argument to interrupt handler
 */
void grspwrouter_isr_register(grspwrouter_dev *dev, uint32_t err_mask, uint32_t port_mask, grspwrouter_isr_func isr, void * arg);

/**
 * @brief Enable timeouts
 *
 * Enable a timeout for packets transfers (overrun and underrun) and auto-disconnect per port.
 *
 * An overrun timeout occurs when the input port has data available but
 * the output port(s) can not accept data fast enough. An underrun timeout occurs when the output
 * port(s) can accept more data but the input port can not provide data fast enough.
 * The timeout can be set to between 100μs - 6.5s.
 *
 * @param[in] dev Valid router device handle
 * @param port Index of a valid SpaceWire port
 * @param timeout The timeout in microseconds (100 - 6553500 in steps of 100)
 * @param overrun Enable for overrun
 * @param underrun Enable for underrun
 * @param autodisconnect Enable for auto disconnect (Only for SpaceWire ports)
 * @retval DRV_OK on success
 * @retval DRV_INVAL if trying to enable auto disconnect on non-SpaceWire port, or if invalid port, or if the timeout value is too big
 */
drvret grspwrouter_port_timeout(grspwrouter_dev * dev, uint8_t port, uint32_t timeout, bool overrun, bool underrun, bool autodisconnect);

/**
 * @brief Abort packet on time/interrupt code
 *
 * Configure the port to abort the current packet if a time or interrupt code
 * with the specified value is received.
 *
 * @param[in] dev Valid router device handle
 * @param port Index of a valid SpaceWire port
 * @param enable Enable packet truncation
 * @param value The value that can cause truncation
 * @param mask Mask for the value
 *
 * @retval DRV_OK on success
 * @retval DRV_INVAL if invalid port
 */
drvret grspwrouter_port_code_truncation(grspwrouter_dev * dev, uint8_t port, bool enable, uint8_t value, uint8_t mask);

/**
 * @brief Set the maximum length of packets
 *
 * If an incoming packets is larger it will be truncated and marked with an EEP.
 * Use the length 0 to accept any length.
 *
 * @param[in] dev Valid router device handle
 * @param port Index of a valid SpaceWire port
 * @param length The maximum length of the packet or 0 to disable
 *
 * @retval DRV_OK on success
 * @retval DRV_INVAL if invalid port
 */
drvret grspwrouter_port_max_length(grspwrouter_dev * dev, uint8_t port, uint32_t length);

/**
 * @brief Read the credit counters for the port
 *
 * Returns the current credit counters for the SpaceWire port. Can not be
 * used on an AMBA port.
 *
 * @param[in] dev Valid router device handle
 * @param port Index of a valid SpaceWire port
 * @param[out] in Incoming credit
 * @param[out] out Outgoing credit
 *
 * @retval DRV_OK on success
 * @retval DRV_INVAL if invalid SpaceWire port
 */
drvret grspwrouter_port_cred(grspwrouter_dev * dev, uint8_t port, uint8_t *in, uint8_t *out);

/**
 * @brief Set the SpaceWire Plug-and-Play information
 *
 * Sets the serial number, vendor id, and product id that is presented when
 * accessing this device using the SpaceWire Plug-and-Play protocol. Bits
 * 3:0 of the serial number can be set using the INSTANCEID[7:0] signal. Use
 * ``keep_instance_id`` to preserve this part of the serial number.
 *
 * @param[in] dev Valid router device handle
 * @param vendor_id Custom vendor id
 * @param product_id Custom product id
 * @param serial Custom serial number
 * @param keep_instance_id Use reset value for bits 3:0 of serial number
 */
void grspwrouter_pnp_set(grspwrouter_dev *dev, uint16_t vendor_id, uint16_t product_id, uint32_t serial, bool keep_instance_id);

/**
 * @brief Interrupt service routine for SpaceWire Router driver
 *
 * Needs to be registered to the OS interrupt handler and provide the
 * device handle for an initialized device. Should only be called from
 * the interrupt handler.
 *
 * @param[in] dev Valid router device handle
 */
void grspwrouter_isr(grspwrouter_dev *dev);

/**
 * @brief Return the current mapped address
 *
 * Returns routers current mapped address. GR716b only.
 *
 * @param[in] dev Valid router device handle
 * @return Current mapped address
 */
uint8_t gr716b_grspwrouter_mapped_adr_get(grspwrouter_dev *dev);

/**
 * @brief Resets the current mapped adress
 *
 * Reset the currently mapped address on GR716B. The currently mapped address
 * needs to be reset before a new address can be mapped.
 *
 * @param[in] dev Valid router device handle
 *
 * @retval DRV_OK on success
 */
drvret gr716b_grspwrouter_mapped_adr_reset(grspwrouter_dev *dev);