/*
 * Copyright (c) 2023, Cobham Gaisler AB
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef DRV_NAND_H
#define DRV_NAND_H

/* Access library for GRLIB NANDFCTRL2 */

#include <stdint.h>
#include <drv/auto.h>

enum {
  NAND_OK          =  0,
  NAND_NOCHIP,
  NAND_TIMEOUT,
  NAND_BADCONFIG, /* Config record is not compatible with hardware */
  NAND_EIO,       /* Error when performing I/O operation */
  NAND_BADARG,    /* Function argument not compatible with hardware */
  NAND_EFAIL       = -1234,
};

struct nand_devcfg;
struct nand_priv;

/* return: number of devices registered */
int nand_autoinit(void);
/* Register one device */
int nand_register(struct nand_devcfg *devcfg);
/* Register an array of devices */
int nand_init(struct nand_devcfg *devcfgs[]);
/* Number of APBUART devices registered */
int nand_dev_count(void);

struct nand_priv *nand_open(int dev_no);
int nand_close(struct nand_priv *priv);

/* All values are in ns */
struct nf2_timing {
  int tcs;  int tww;
  int trhw; int trr;  int twb;
  int tadl; int twhr; int tccs;
  int twh;  int twp;  int treh; int trp;
};

extern const struct nf2_timing NAND_MODE0;
extern const struct nf2_timing NAND_MODE1;
extern const struct nf2_timing NAND_MODE2;
extern const struct nf2_timing NAND_MODE3;
extern const struct nf2_timing NAND_MODE4;
extern const struct nf2_timing NAND_MODE5;

struct nand_config {
  unsigned long long phy_freq_hz;
  int msel;
  /* for each chip enable ("target") */
  struct nand_config_map {
    int en;    /* use this chip enable */
    int chsel; /* mask of channels to operate on */
    int rbsel; /* mask of ready signals used */
    int sefi;
  } map[64];
};

/* 3 raw parameter pages, ONFI endian */
struct onfi_parameter_page_x3 {
  struct {
    unsigned char raw[256];
  } copy[3];
};

/* Parsed parameter page */
struct onfi_upp {
  struct {
    char sign[5];
    int rev;
    int feat;
    int optcmd;
    int advopt;
    int extpplen;
    int npp;
  } info;
  struct {
    char man[13];
    char model[21];
    int jedec_mid;
    int date;
  } man;
  struct {
    int data_bytes_per_page;
    int spare_bytes_per_page;
    int pages_per_block;
    int blocks_per_lun;
    int luns;
    int col_address_cycles;
    int row_address_cycles;
    int bits_per_cell;
    int plane_address_bits;
  } mem;
  struct {
    unsigned sdr_modes;
    unsigned nvddr_modes;
    unsigned nvddr2_modes;
  } ele;
  struct {
    int rev;
    int crc;
  } ven;
};

/* These are actual numbers (1 already added) */
struct nand_ucap {
  int memsup;
  int ft;
  int prbbm;
  int rnd;
  int necc;
  int nrb;      /* RB_N and SEFI */
  int nchan;    /* buses: IO0-7, DQS, OE, CLE, ALE, RE_N, WE_N, WP_N */
  int ntarg;    /* number of chip selects (CE_N) */
  struct {
    int cap;
    int chunk;
    int gf;
  } ecc[2];     /* necc */
  struct nand_ucap_mem {
    int sel;      /* index into ecc[] */
    int spare;    /* number of bytes */
    int data;     /* number of bytes*/
  } mem[3];     /* valid index is 0 to memsup-1 */
};

/*
 * Retreive a copy of the controller capabilities
 *
 * ucap: capabilities will be written here.
 * return: NAND_OK
 */
int nand_get_cap(
  struct nand_priv *priv,
  struct nand_ucap *ucap
);

/* This will copy config to the driver priv and program MODE0 timing */
int nand_config(
  struct nand_priv *priv,
  const struct nand_config *config
);

/* Print info about the controller configuration. Available after nand_open(). */
int nand_print_info(
  struct nand_priv *priv
);

int nand_set_timing(
  struct nand_priv *priv,
  const struct nf2_timing *v,
  int edo
);

/* Update internal knowledge about block sizes, etc. */
int nand_set_geom(
  struct nand_priv *priv,
  int pages_per_block,
  int blocks_per_lun,
  int luns
);

long nand_get_row(
  struct nand_priv *priv,
  int lun,
  int block,
  int page
);

int nand_cmd_reset(
  struct nand_priv *priv,
  int cs
);

int nand_cmd_read_status(
  struct nand_priv *priv,
  int cs,
  int *status
);

int nand_cmd_read_id(
  struct nand_priv *priv,
  int cs,
  int addr,
  unsigned char *buf,
  int len
);

/* Always reads the 3 first copies, each 256 byte, of parameter page */
int nand_cmd_read_parameter_page(
  struct nand_priv *priv,
  int cs,
  struct onfi_parameter_page_x3 *buf
);

int nand_parse_parameter_page(
  struct nand_priv *priv,
  const struct onfi_parameter_page_x3 *raw,
  struct onfi_upp *u
);

/* read raw without randomization */
int nand_cmd_read_spare(
  struct nand_priv *priv,
  int cs,
  unsigned char *buf,
  int *len,
  int lun,
  int block,
  int page,
  int offset
);

/* nominal read full page with NANDFCTRL2 ECC and randomization */
int nand_cmd_read(
  struct nand_priv *priv,
  int cs,
  unsigned char *buf,
  int lun,
  int block,
  int page
);

/* nominal program full page with NANDFCTRL2 ECC and randomization */
int nand_cmd_page_program(
  struct nand_priv *priv,
  int cs,
  unsigned char *buf,
  int lun,
  int block,
  int page
);

struct nf2_desc {
  uint32_t dcmd;        /* 00 */
  uint32_t dtarsel[2];  /* 04..08 */
  uint32_t dchsel;      /* 0c */
  uint32_t drbsel;      /* 10 */
  uint32_t drow;
  uint32_t dcol;
  uint32_t dstat;
  uint32_t deccstat;    /* 20 */
  uint32_t x24;
  uint32_t x28;
  uint32_t x2c;
  uint32_t ddpl;        /* 30 data low */
  uint32_t ddph;        /* 34 data high */
  uint32_t nextl;       /* 38 Descriptor linked list pointer low */
  uint32_t nexth;       /* 3c Descriptor linked list pointer high */
};

struct nand_dma_info {
  struct nf2_desc desc;
  unsigned char *buf;
  struct nand_dma_info *next;
  int cs;
  int lun;
  int block;
  int page;
};

int nand_cmd_dma_page_program(
  struct nand_priv *priv,
  struct nand_dma_info *info /* linked list of info structures */
);

int nand_cmd_dma_read(
  struct nand_priv *priv,
  struct nand_dma_info *info /* linked list of info structures */
);

int nand_cmd_block_erase(
  struct nand_priv *priv,
  int cs,
  int lun,
  int block
);


/** PRIVATE **/
/* Driver private structure. Shall never be referenced by user. */
struct nf2_regs;

struct nand_priv {
  volatile struct nf2_regs *regs;
  int                intnum;
  int                bigendian;
  /* current user-supplied chip config */
  struct nand_config cfg;
  /* raw copy of cababilities registers */
  uint32_t           cap[5];
  /* human-readable cababilities */
  struct nand_ucap   ucap;
  /* Geometry of attached memory */
  struct {
    int                pages_per_block;
    int                blocks_per_lun;
    int                luns;
  } geom;
  uint8_t              open;
};

struct nand_devcfg {
  struct drv_devreg regs;
  struct nand_priv priv;
};

#endif

